#ifndef _CRT_RAND_S
#define _CRT_RAND_S
#endif
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include "../include/lmc.h"

wchar_t g_lmcsRelativePath[MAX_PATH];
wchar_t* g_lmclpInsertionPoint;

HANDLE LMCOpenFile(LPCWSTR lpFileName)
{
	return CreateFile(lpFileName, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
}

DWORD LMCReadDictionary(HANDLE hFile, LPDICTIONARY_ENTRY* lpDictionaryBuffer, LPDWORD lpEntriesFound)
{
	DWORD dwLMCVersion;
	DWORD dwDictionarySeed;
	DWORD dwDictionaryLength;
	LPBYTE lpDictionaryRaw;
	LPBYTE lpDictionaryRawPointer;
	LPDICTIONARY_ENTRY lpDictionary;
	DWORD dwEntryCount;
	DWORD dwCurrentEntry;

	SetFilePointer(hFile, 0, NULL, FILE_BEGIN);

	// マジックナンバー確認
	if (LMCAuxReadDWORD(hFile) != 0x434D4C58)
	{
		SetLastError(ERROR_INVALID_FUNCTION);
		return 0;
	}
	if ((dwLMCVersion = LMCAuxReadDWORD(hFile)) != 0x00000108)
	{
		SetLastError(ERROR_NOT_SUPPORTED);
		return 0;
	}

	// ヘッダー読み込み
	dwDictionarySeed = LMCAuxReadDWORD(hFile);
	dwDictionaryLength = LMCAuxReadDWORD(hFile);
	LMCXORBytes(dwDictionarySeed & 0x1104E108, &dwDictionaryLength, 4);

	// 生のディクショナリを読み込んで…
	lpDictionaryRaw = LMCAuxReadBuffer(hFile, dwDictionaryLength);

	// デコードして…
	LMCXORBytes(dwDictionarySeed, lpDictionaryRaw, dwDictionaryLength);
	SecureZeroMemory(&dwDictionarySeed, 4); // これはもういらない...
	lpDictionaryRawPointer = lpDictionaryRaw;
	dwEntryCount = *((DWORD*)lpDictionaryRawPointer);
	lpDictionaryRawPointer += 4;

	// 本当のディクショナリを作成する
	lpDictionary = LMCCreateBuffer(dwEntryCount * sizeof(DICTIONARY_ENTRY));

	// 項目を1つずつ読み込む
	for (dwCurrentEntry = 0; dwCurrentEntry < dwEntryCount; dwCurrentEntry++)
	{
		CopyMemory(&(lpDictionary[dwCurrentEntry].dwEntryNameSize), lpDictionaryRawPointer, 4);
		lpDictionaryRawPointer += 4;

		lpDictionary[dwCurrentEntry].lpEntryName = LMCCreateBuffer(lpDictionary[dwCurrentEntry].dwEntryNameSize * sizeof(wchar_t));
		CopyMemory(lpDictionary[dwCurrentEntry].lpEntryName, lpDictionaryRawPointer, lpDictionary[dwCurrentEntry].dwEntryNameSize * sizeof(wchar_t));
		lpDictionaryRawPointer += lpDictionary[dwCurrentEntry].dwEntryNameSize * sizeof(wchar_t);
		
		CopyMemory(&(lpDictionary[dwCurrentEntry].dwOffset), lpDictionaryRawPointer, 4);
		lpDictionaryRawPointer += 4;
		CopyMemory(&(lpDictionary[dwCurrentEntry].dwSize), lpDictionaryRawPointer, 4);
		lpDictionaryRawPointer += 4;
		CopyMemory(&(lpDictionary[dwCurrentEntry].dwSeed), lpDictionaryRawPointer, 4);
		lpDictionaryRawPointer += 4;
	}

	// メモリーリリース
	LMCFreeBuffer(lpDictionaryRaw);

	*lpDictionaryBuffer = lpDictionary;
	*lpEntriesFound = dwEntryCount;
	return 1;
}

DWORD LMCExtractToFile(HANDLE hFile, LPDICTIONARY_ENTRY lpEntry, LPCWSTR lpDestinationPath)
{
	LPVOID lpBuffer;
	HANDLE hOutputFile;
	if (lpEntry == NULL)
		return 0;
	LMCExtractToMemory(hFile, lpEntry, &lpBuffer);
	hOutputFile = CreateFile(lpDestinationPath, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	LMCAuxWriteBuffer(hOutputFile, lpBuffer, lpEntry->dwSize);
	CloseHandle(hOutputFile);
	LMCFreeBuffer(lpBuffer);
	return 1;
}

DWORD LMCExtractToMemory(HANDLE hFile, LPDICTIONARY_ENTRY lpEntry, LPVOID* lpDestinationBuffer)
{
	LPVOID lpBuffer;
	*lpDestinationBuffer = NULL;
	if (lpEntry == NULL)
		return 0;
	SetFilePointer(hFile, lpEntry->dwOffset, NULL, FILE_BEGIN);
	lpBuffer = LMCAuxReadBuffer(hFile, lpEntry->dwSize);
	LMCXORBytes(lpEntry->dwSeed, lpBuffer, lpEntry->dwSize);
	*lpDestinationBuffer = lpBuffer;
	return 1;
}

BOOL LMCCloseFile(HANDLE hFile, LPDICTIONARY_ENTRY lpDictionaryBuffer, DWORD dwEntryCount)
{
	DWORD dwCurrentEntry;
	for (dwCurrentEntry = 0; dwCurrentEntry < dwEntryCount; dwCurrentEntry++)
		LMCFreeBuffer(lpDictionaryBuffer[dwCurrentEntry].lpEntryName);
	LMCFreeBuffer(lpDictionaryBuffer);
	return CloseHandle(hFile);
}

LPDICTIONARY_ENTRY LMCFindFile(LPDICTIONARY_ENTRY lpDictionary, DWORD dwEntryCount, LPCWSTR lpFileName)
{
	DWORD i;
	for (i = 0; i < dwEntryCount; i++)
	{
		if (wcscmp(lpDictionary[i].lpEntryName, lpFileName) == 0)
			return &(lpDictionary[i]);
	}
	return NULL;
}

LPFILE_LIST LMCCreateFileList(LPCWSTR lpBasePath)
{
	LPFILE_LIST lpFileList;
	wcscpy_s(g_lmcsRelativePath, MAX_PATH, lpBasePath);
	for (g_lmclpInsertionPoint = g_lmcsRelativePath; ; g_lmclpInsertionPoint++)
	{
		if (*g_lmclpInsertionPoint == L'\r' || *g_lmclpInsertionPoint == L'\n')
		{
			if (*g_lmclpInsertionPoint != L'\\')
			{
				*g_lmclpInsertionPoint = L'\\';
				g_lmclpInsertionPoint++;
			}
			break;
		}
	}
	lpFileList = LMCCreateBuffer(sizeof(FILE_LIST));
	lpFileList->lpNextEntry = NULL;
	return lpFileList;
}

BOOL LMCAddFile(LPFILE_LIST lpFileList, LPCWSTR lpPath)
{
	int nBufferLength;
	BOOL bIsFirstEntry;
	LPFILE_LIST_ENTRY lpCurrentEntry;
	LPFILE_LIST_ENTRY lpNewEntry;
	wchar_t* p;
	if (lpFileList != NULL)
	{
		bIsFirstEntry = lpFileList->lpPath == NULL;
		if (bIsFirstEntry) // 一番最初のを直接に読み込む
			lpNewEntry = lpFileList;
		else // 他のはバッファーを作成する
			lpNewEntry = LMCCreateBuffer(sizeof(FILE_LIST_ENTRY));
		lpNewEntry->lpNextEntry = NULL;
		nBufferLength = wcslen(lpPath) + 1;
		lpNewEntry->lpPath = LMCCreateBuffer(nBufferLength * sizeof(wchar_t));
		wcscpy_s(lpNewEntry->lpPath, nBufferLength, lpPath);
		for (p = lpNewEntry->lpPath; *p; p++)
			if (*p == L'\r' || *p == L'\n')
				*p = 0;
		if (!bIsFirstEntry) // 項目を追加する
		{
			lpCurrentEntry = lpFileList;
			while (lpCurrentEntry->lpNextEntry != NULL)
				lpCurrentEntry = lpCurrentEntry->lpNextEntry;
			lpCurrentEntry->lpNextEntry = lpNewEntry;
		}
		return TRUE;
	}
	return FALSE;
}

BOOL LMCWriteFile(LPCWSTR lpFileName, LPFILE_LIST lpFileList)
{
	LPFILE_LIST_ENTRY lpCurrentEntry;
	LPDICTIONARY_ENTRY lpDictionary;
	WIN32_FILE_ATTRIBUTE_DATA dFileData;
	int nDictionaryEntryCount;
	int dwCurrentEntry;
	int nTotalDictionaryLength;
	int nCurrentOffset;
	HANDLE hOutputFile;
	DWORD dwDictionarySeed;
	LPBYTE lpDictionaryRaw;
	LPBYTE lpDictionaryRawPointer;
	LPVOID lpCurrentFile;
	int e;

	if (lpFileList != NULL)
	{
		lpCurrentEntry = lpFileList;
		nDictionaryEntryCount = 0;
		while (lpCurrentEntry != NULL)
		{
			lpCurrentEntry = lpCurrentEntry->lpNextEntry;
			nDictionaryEntryCount++;
		}
		lpDictionary = LMCCreateBuffer(nDictionaryEntryCount * sizeof(DICTIONARY_ENTRY));
		lpCurrentEntry = lpFileList;
		dwCurrentEntry = 0;

        // ポインターサイズを消す。実際のサイズは後で入れるから
        // そして、エントリー数の為に、4バイトを開ける
		nTotalDictionaryLength = nDictionaryEntryCount * (sizeof(DICTIONARY_ENTRY) - sizeof(LPWSTR)) + 4;
		while (lpCurrentEntry != NULL)
		{
			wcscpy_s(g_lmclpInsertionPoint, MAX_PATH, lpCurrentEntry->lpPath); // 知ってる、知ってる
			lpDictionary[dwCurrentEntry].dwEntryNameSize = wcslen(lpCurrentEntry->lpPath) + 1;
			nTotalDictionaryLength += lpDictionary[dwCurrentEntry].dwEntryNameSize * sizeof(wchar_t);
			lpDictionary[dwCurrentEntry].lpEntryName = lpCurrentEntry->lpPath;
			GetFileAttributesEx(g_lmcsRelativePath, GetFileExInfoStandard, &dFileData);
			lpDictionary[dwCurrentEntry].dwSize = dFileData.nFileSizeLow;
			// ヘッダーサイズが分かったら、lpDictionary[dwCurrentEntry].dwOffsetを追加する
			lpDictionary[dwCurrentEntry].dwSeed = LMCCreateSeed();
			lpCurrentEntry = lpCurrentEntry->lpNextEntry;
			dwCurrentEntry++;
		}

		// nTotalDictionaryLengthに、ヘッダーの全体のサイズが入っている。
        // マジックナンバーでプラス４，シードでプラス４、ディクショナリサイズでプラス４だと、すべての
        // オフセットを計算することができる
		nCurrentOffset = 16 + nTotalDictionaryLength;
		for (dwCurrentEntry = 0; dwCurrentEntry < nDictionaryEntryCount; dwCurrentEntry++)
		{
			lpDictionary[dwCurrentEntry].dwOffset = nCurrentOffset;
			nCurrentOffset += lpDictionary[dwCurrentEntry].dwSize;
		}

		// 本番のディクショナリを作成する
		lpDictionaryRaw = LMCCreateBuffer(nTotalDictionaryLength);
		lpDictionaryRawPointer = lpDictionaryRaw;
		*((DWORD*)lpDictionaryRawPointer) = nDictionaryEntryCount;
		lpDictionaryRawPointer += 4;
		for (dwCurrentEntry = 0; dwCurrentEntry < nDictionaryEntryCount; dwCurrentEntry++)
		{
			CopyMemory(lpDictionaryRawPointer, &(lpDictionary[dwCurrentEntry].dwEntryNameSize), 4);
			lpDictionaryRawPointer += 4;
			CopyMemory(lpDictionaryRawPointer,   lpDictionary[dwCurrentEntry].lpEntryName, lpDictionary[dwCurrentEntry].dwEntryNameSize * sizeof(wchar_t));
			lpDictionaryRawPointer += lpDictionary[dwCurrentEntry].dwEntryNameSize * sizeof(wchar_t);
			CopyMemory(lpDictionaryRawPointer, &(lpDictionary[dwCurrentEntry].dwOffset), 4);
			lpDictionaryRawPointer += 4;
			CopyMemory(lpDictionaryRawPointer, &(lpDictionary[dwCurrentEntry].dwSize), 4);
			lpDictionaryRawPointer += 4;
			CopyMemory(lpDictionaryRawPointer, &(lpDictionary[dwCurrentEntry].dwSeed), 4);
			lpDictionaryRawPointer += 4;
		}
		assert(lpDictionaryRaw + nTotalDictionaryLength == lpDictionaryRawPointer);
		dwDictionarySeed = 0x999bcb9c; // LMCCreateSeed();
		LMCXORBytes(dwDictionarySeed, lpDictionaryRaw, nTotalDictionaryLength);

		// よっしゃ、ファイル作成開始！
		hOutputFile = CreateFile(lpFileName, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		if (hOutputFile == INVALID_HANDLE_VALUE)
		{
			e = GetLastError();
			LMCFreeBuffer(lpDictionary);
			return FALSE;
		}

		// ヘッダー
		LMCAuxWriteDWORD(hOutputFile, 0x434D4C58); // マジックナンバー
		LMCAuxWriteDWORD(hOutputFile, 0x00000108); // バージョン
		LMCAuxWriteDWORD(hOutputFile, dwDictionarySeed);
		LMCXORBytes(dwDictionarySeed & 0x1104E108, &nTotalDictionaryLength, 4);
		LMCAuxWriteDWORD(hOutputFile, nTotalDictionaryLength);
		LMCXORBytes(dwDictionarySeed & 0x1104E108, &nTotalDictionaryLength, 4);
		nCurrentOffset = 16; // これをリユースする

		// ディクショナリ
		LMCAuxWriteBuffer(hOutputFile, lpDictionaryRaw, nTotalDictionaryLength);
		LMCFreeBuffer(lpDictionaryRaw);
		nCurrentOffset += nTotalDictionaryLength;

		// すべてのファイル
		for (dwCurrentEntry = 0; dwCurrentEntry < nDictionaryEntryCount; dwCurrentEntry++)
		{
			assert(nCurrentOffset == lpDictionary[dwCurrentEntry].dwOffset);
			// ファイルを読み雲
			wcscpy_s(g_lmclpInsertionPoint, MAX_PATH, lpDictionary[dwCurrentEntry].lpEntryName); // <-- I know, I know...
			lpCurrentFile = LMCAuxReadFile(g_lmcsRelativePath, lpDictionary[dwCurrentEntry].dwSize);
			if (lpCurrentFile == INVALID_HANDLE_VALUE)
				return FALSE;
			// XOR化
			LMCXORBytes(lpDictionary[dwCurrentEntry].dwSeed, lpCurrentFile, lpDictionary[dwCurrentEntry].dwSize);
			// バッファー書き込み
			LMCAuxWriteBuffer(hOutputFile, lpCurrentFile, lpDictionary[dwCurrentEntry].dwSize);
			// そして開放
			LMCFreeBuffer(lpCurrentFile);
			nCurrentOffset += lpDictionary[dwCurrentEntry].dwSize;
		}
		
		// 最後にファイルを閉じる
		CloseHandle(hOutputFile);

		// バッファーを開放する
		LMCFreeBuffer(lpDictionary);
		return TRUE;
	}
	return FALSE;
}

BOOL LMCFreeFileList(LPFILE_LIST lpFileList)
{
	if (lpFileList != NULL)
	{
		LMCFreeFileList(lpFileList->lpNextEntry);
		LMCFreeBuffer(lpFileList->lpPath);
		LMCFreeBuffer(lpFileList);
	}
	return TRUE;
}

DWORD LMCCreateSeed()
{
	DWORD dwValue;
	rand_s(&dwValue);
	return dwValue;
}

DWORD LMCXORBytes(DWORD dwSeed, LPVOID lpBuffer, DWORD dwBufferLength)
{
	DWORD i;
	srand(dwSeed);
	for (i = 0; i < dwBufferLength; i++)
		((char*)lpBuffer)[i] ^= rand() % 0xff;
	return 1;
}

int LMCVLIntToInt(int value, int* lpBytesUsed)
{
    // .netストリングサイズからの変換
	int v;
	int m;
	char* p;
	int b;
	v = 0;
	m = 1;
	b = 1;
	p = (char*)(&value);
	while (*p & 0x80)
	{
		v += (*p & 0x7f) * m;
		m *= 0x80;
		p++;
		b++;
	}
	v += *p * m;
	if (lpBytesUsed != 0)
		*lpBytesUsed = b;
	return v;
}

int LMCIntToVLInt(int value, int* lpBytesUsed)
{
    // .netストリングシアズへの変換
	int v;
	char* p;
	int b;
	v = 0;
	p = (char*)(&v);
	b = 1;
	while (value >= 0x80)
	{
		*p = ((char*)(&value))[0] | 0x80;
		value = value >> 7;
		p++;
		b++;
	}
	*p = ((char*)(&value))[0];
	if (lpBytesUsed != 0)
		*lpBytesUsed = b;
	return v;
}

LPVOID LMCCreateBuffer(DWORD dwSize)
{
	return HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, dwSize);
}

DWORD LMCFreeBuffer(LPVOID lpBuffer)
{
	DWORD dwSize;
	HANDLE hHeap;

	hHeap = GetProcessHeap();
	dwSize = HeapSize(hHeap, 0, lpBuffer);
	SecureZeroMemory(lpBuffer, dwSize);
	HeapFree(hHeap, 0, lpBuffer);
	return 1;
}

BOOL LMCAuxWriteDWORD(HANDLE hFile, DWORD dwValue)
{
	return LMCAuxWriteBuffer(hFile, &dwValue, 4);
}

BOOL LMCAuxWriteBuffer(HANDLE hFile, LPVOID lpBuffer, int nBytesToWrite)
{
	DWORD dwBytesWritten;
	return WriteFile(hFile, lpBuffer, nBytesToWrite, &dwBytesWritten, NULL);
}

DWORD LMCAuxReadDWORD(HANDLE hFile)
{
	DWORD dwValue;
	DWORD dwBytesRead;
	ReadFile(hFile, &dwValue, 4, &dwBytesRead, NULL);
	return dwValue;
}

// 読み込み先がバッファーを開放する
LPVOID LMCAuxReadBuffer(HANDLE hFile, int nBytesToRead)
{
	LPVOID lpBuffer;
	DWORD dwBytesRead;
	lpBuffer = LMCCreateBuffer(nBytesToRead);
	ReadFile(hFile, lpBuffer, nBytesToRead, &dwBytesRead, NULL);
	return lpBuffer;
}

LPVOID LMCAuxReadFile(LPCWSTR lpFileName, int nBytesToRead)
{
	HANDLE hFile;
	LPVOID lpBuffer;
	hFile = CreateFile(lpFileName, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFile == INVALID_HANDLE_VALUE)
		return INVALID_HANDLE_VALUE;
	lpBuffer = LMCAuxReadBuffer(hFile, nBytesToRead);
	CloseHandle(hFile);
	return lpBuffer;
}
